"""
TODO : documentation ...
"""
import numpy as np
import scipy.sparse as sp

from marge3d.params import DaitcheParameters

class NumericalSolver:

    def __init__(self, x, w, velocity_field, Nt, order, particle_density, fluid_density, particle_radius,
                       kinematic_viscosity, time_scale, char_vel):

        self.w       = np.copy(w)       # velocity difference
        self.x       = np.copy(x)       # particle position
        self.p       = DaitcheParameters(particle_density, fluid_density,
                            particle_radius, kinematic_viscosity, time_scale, char_vel)


        self.vel     = velocity_field

        '''
        if self.vel.limits == True:
            if (x[0] > self.vel.x_right or x[0] < self.vel.x_left or x[1] > self.vel.y_up or x[1] < self.vel.y_down):
                  raise Exception("Particle's initial position is outside the spatial domain")
        '''

        if order == 1:
            self.calc_alpha_mat(Nt)
        elif order == 2:
            self.euler_nodes  = 151 # This number could be increased to increase accuracy
            self.calc_alpha_mat(self.euler_nodes)
            self.calc_beta_mat(Nt)
        elif order == 3:
            self.euler_nodes  = 151 # This number could be increased to increase accuracy
            self.calc_alpha_mat(self.euler_nodes)
            self.calc_beta_mat(self.euler_nodes)
            self.calc_gamma_mat(Nt)
        else:
            raise("Requested order for Daitche's method not available.")


    def calculate_G(self, w1, w2, w3, x, y, z, t):

        coeff              = self.p.R - 1.0

        #u1, u2, u3         = self.vel.get_velocity(x, y, z, t)
        u1, u2, u3         = self.vel.get_velocity(x, y, z, t)

        u1x, u1y, u1z, u2x, u2y, u2z, u3x, u3y, u3z  = self.vel.get_gradient(x, y, z, t)

        u1t, u2t, u3t      = self.vel.get_dudt(x, y, z, t)

        G1                 = coeff * u1t + (coeff * u1 - w1) * u1x + \
                                      (coeff * u2 - w2) * u1y + \
                                        (coeff * u3 - w3) * u1z

        G2                 = coeff * u2t + (coeff * u1 - w1) * u2x + \
                                      (coeff * u2 - w2) * u2y  + \
                                        (coeff * u3 - w3) * u2z

        G3                 = coeff * u3t + (coeff * u1 - w1) * u3x + \
                                       (coeff * u2 - w2) * u3y  + \
                                         (coeff * u3 - w3) * u3z
        return G1, G2, G3



    def alpha_jn(self, j, n):
        if n == 0:
            return 0.0
        elif j == 0:
            return 4.0 / 3.0
        elif j != n:
            return ((j - 1.0)**1.5 + (j + 1.0)**1.5 - 2.0 * j**1.5) * 4.0 / 3.0
        else:
            return ((n - 1.0)**1.5 - n**1.5 + np.sqrt(n)*3.0/2.0) * 4.0/3.0



    def alpha_v(self, N):

        AlphaSubst_v = np.array([])
        for j in range(0, N+1):
            AlphaSubst_v = np.append(AlphaSubst_v,
                              self.alpha_jn(j+1, N+1) -\
                                  self.alpha_jn(j, N))

        return AlphaSubst_v


    def calc_alpha_mat(self, N):
        for nn in range(0, N-1):
            if nn == 0:
                alpha_mat     = sp.csr_matrix(self.alpha_v(nn),
                                              shape=(1, N))
            else:
                alpha_v       = sp.csr_matrix(self.alpha_v(nn),
                                              shape=(1, N))
                alpha_mat     = sp.vstack([alpha_mat, alpha_v])
        self.alpha_mat = alpha_mat


    def Euler(self, t_v, flag=False):
        assert len(t_v) >= 2, "This time grid cannot be used for time stepping."

        h     = t_v[1] - t_v[0]

        x_v   = np.array([self.x[0]])
        y_v   = np.array([self.x[1]])
        z_v   = np.array([self.x[2]])

        u10, u20, u30 = self.vel.get_velocity(self.x[0], self.x[1], self.x[2], t_v[0])
        w1_v  = np.array([self.w[0]])
        w2_v  = np.array([self.w[1]])
        w3_v  = np.array([self.w[2]])

        xi    = self.p.R * np.sqrt(3 / self.p.S) * np.sqrt(h / np.pi)

        for nn in range(0, len(t_v)-1):

            # TODO : why make it a csr matrix if you transform it back to a dense array ?
            Sum_w1        = np.dot(self.alpha_mat.toarray()[nn, :nn+1], w1_v)
            Sum_w2        = np.dot(self.alpha_mat.toarray()[nn, :nn+1], w2_v)
            Sum_w3        = np.dot(self.alpha_mat.toarray()[nn, :nn+1], w3_v)

            if nn == 0:
                x_np1     = x_v[nn] + h * (w1_v[0] + u10)
                y_np1     = y_v[nn] + h * (w2_v[0] + u20)
                z_np1     = z_v[nn] + h * (w3_v[0] + u30)
            else:
                u1_n, u2_n, u3_n  = self.vel.get_velocity(x_v[nn], y_v[nn], z_v[nn], t_v[nn])

                x_np1     = x_v[nn] + h * (w1_v[0] + u1_n)
                y_np1     = y_v[nn] + h * (w2_v[0] + u2_n)
                z_np1     = z_v[nn] + h * (w3_v[0] + u3_n)

            G1, G2, G3    = self.calculate_G(w1_v[0], w2_v[0], w3_v[0],
                                        x_v[nn], y_v[nn], z_v[nn],
                                        t_v[nn])

            Gw1_n         = G1 - (self.p.R / self.p.S) * w1_v[0] - (1 - self.p.R) * 0.0
            Gw2_n         = G2 - (self.p.R / self.p.S) * w2_v[0] - (1 - self.p.R) * 0.0
            Gw3_n         = G3 - (self.p.R / self.p.S) * w3_v[0] - (1 - self.p.R) * self.p.g


            w1_np1        = ( w1_v[0] + h * Gw1_n - xi * Sum_w1 ) / \
                            (1.0 + xi * self.alpha_jn(0, nn+1))
            w2_np1        = ( w2_v[0] + h * Gw2_n - xi * Sum_w2 ) / \
                            (1.0 + xi * self.alpha_jn(0, nn+1))
            w3_np1        = ( w3_v[0] + h * Gw3_n - xi * Sum_w3 ) / \
                            (1.0 + xi * self.alpha_jn(0, nn+1))


            x_v        = np.append(x_v, x_np1)
            y_v        = np.append(y_v, y_np1)
            z_v        = np.append(z_v, z_np1)
            w1_v       = np.append(w1_np1, w1_v)
            w2_v       = np.append(w2_np1, w2_v)
            w3_v       = np.append(w3_np1, w3_v)

        pos_vec_x      = x_v
        pos_vec_y      = y_v
        pos_vec_z      = z_v
        w_vec          = np.transpose(np.array([np.flip(w1_v), np.flip(w2_v), np.flip(w3_v)]))

        if flag == True:
            self.pos_vec_x = np.copy(pos_vec_x)
            self.pos_vec_y = np.copy(pos_vec_y)
            self.pos_vec_z = np.copy(pos_vec_z)
            self.w_vec     = np.copy(w_vec)

        return pos_vec_x, pos_vec_y, pos_vec_z, w_vec


    def beta_jn(self, j, n):

        if n == 0:
            return 0.0

        elif n == 1:
            return self.alpha_jn(j, n)

        elif n == 2:
            if j == 0:
                return 12.0 * np.sqrt(2.0) / 15.0
            elif j == 1:
                return 16.0 * np.sqrt(2.0) / 15.0
            elif j == 2:
                return 2.0 * np.sqrt(2.0) / 15.0

        elif n == 3:
            if j == 0:
                return 4.0 * np.sqrt(2.0) / 5.0
            elif j == 1:
                return (14.0 * np.sqrt(3.0) - 12.0 * np.sqrt(2.0)) / 5.0
            elif j == 2:
                return (12.0 * np.sqrt(2.0) - 8.0 * np.sqrt(3.0)) / 5.0
            elif j == 3:
                return (np.sqrt(3.0) - np.sqrt(2.0)) * (4.0 / 5.0)

        else:
            if j == 0:
                return 4.0 * np.sqrt(2.0) / 5.0
            elif j == 1:
                return (14.0 * np.sqrt(3.0) - 12.0 * np.sqrt(2.0)) / 5.0
            elif j == 2:
                return (176.0/15.0) + ( 12.0 * np.sqrt(2.0) - 42.0 * np.sqrt(3.0)) / 5.0
            elif j == n-1:
                return (8.0/15.0) * (-2.0 * n**(5.0/2.0) + \
                              3.0 * (n - 1.0)**(5.0/2.0) - \
                                    (n - 2.0)**(5.0/2.0)) + \
                         (2.0/3.0) * (4.0 * n**(3.0/2.0) - \
                              3.0 * (n - 1.0)**(3.0/2.0) + \
                                    (n - 2.0)**(3.0/2.0))
            elif j == n:
                return (8.0/15.0) * (       n**(5.0/2.0) - \
                                    (n - 1.0)**(5.0/2.0)) + \
                        (2.0/3.0) * (-3.0 * n**(3.0/2.0) + \
                                    (n - 1.0)**(3.0/2.0)) + \
                                    2.0 * np.sqrt(n)
            else:
                return (8.0/15.0) * ((j + 2.0)**(5.0/2.0) - \
                               3.0 * (j + 1.0)**(5.0/2.0) + \
                                       3.0 * j**(5.0/2.0) - \
                                     (j - 1.0)**(5.0/2.0)) + \
                       (2.0/3.0) * (-(j + 2.0)**(3.0/2.0) + \
                               3.0 * (j + 1.0)**(3.0/2.0) - \
                                       3.0 * j**(3.0/2.0) + \
                                     (j - 1.0)**(3.0/2.0))



    def beta_v(self, N):

        BetaSubst_v = np.array([])
        for nn in range(0, N+1):
            BetaSubst_v = np.append(BetaSubst_v,
                              self.beta_jn(nn+1, N+1) -\
                                  self.beta_jn(nn, N))

        return BetaSubst_v



    def calc_beta_mat(self, N):
        for nn in range(0, N-1):
            if nn == 0:
                beta_mat     = sp.csr_matrix(self.beta_v(nn), shape=(1, N))
            else:
                beta_v       = sp.csr_matrix(self.beta_v(nn), shape=(1, N))
                beta_mat     = sp.vstack([beta_mat, beta_v])
        self.beta_mat = beta_mat



    def AdamBashf2(self, t_v, flag=False):
        assert len(t_v) >= 2, "Time grid cannot be used for time stepping."

        h          = t_v[1] - t_v[0]

        # Obtaining first time step with a 1st order method
        t_np1           = np.linspace(t_v[0], t_v[1], self.euler_nodes)
        pos_vec_x, pos_vec_y, pos_vec_z, w_vec = self.Euler(t_np1, flag=False)

        # Calculating the rest of the solution with a 2nd order method

        x_v   = np.array([self.x[0], pos_vec_x[-1]])
        y_v   = np.array([self.x[1], pos_vec_y[-1]])
        z_v   = np.array([self.x[2], pos_vec_z[-1]])

        u10, u20, u30  = self.vel.get_velocity(self.x[0], self.x[1], self.x[2], t_v[0])

        w1_v   = np.array([w_vec[-1,0], self.w[0]])
        w2_v   = np.array([w_vec[-1,1], self.w[1]])
        w3_v   = np.array([w_vec[-1,2], self.w[2]])

        xi    = (self.p.R) * np.sqrt(3 / self.p.S) * np.sqrt(h / np.pi)

        for nn in range(1, len(t_v)-1):
            Sum_w1     = np.dot(self.beta_mat.toarray()[nn, :nn+1], w1_v)
            Sum_w2     = np.dot(self.beta_mat.toarray()[nn, :nn+1], w2_v)
            Sum_w3     = np.dot(self.beta_mat.toarray()[nn, :nn+1], w3_v)

            u1_n, u2_n, u3_n       = self.vel.get_velocity(x_v[nn], y_v[nn], z_v[nn], t_v[nn])
            u1_nm1, u2_nm1, u3_nm1 = self.vel.get_velocity(x_v[nn-1], y_v[nn-1], z_v[nn-1], t_v[nn-1])

            x_np1     = x_v[nn] + (h/2.0) * ( 3.0 * (w1_v[0] + u1_n) - (w1_v[1] + u1_nm1))
            y_np1     = y_v[nn] + (h/2.0) * ( 3.0 * (w2_v[0] + u2_n) - (w2_v[1] + u2_nm1))
            z_np1     = z_v[nn] + (h/2.0) * ( 3.0 * (w3_v[0] + u3_n) - (w3_v[1] + u3_nm1))

            G1_n, G2_n, G3_n  = self.calculate_G(w1_v[0], w2_v[0], w3_v[0],
                                        x_v[nn], y_v[nn], z_v[nn],
                                        t_v[nn])
            G1_nm1, G2_nm1, G3_nm1 = self.calculate_G(w1_v[1], w2_v[1], w3_v[1],
                                        x_v[nn-1], y_v[nn-1], z_v[nn-1],
                                        t_v[nn-1])


            Gw1_n      = G1_n - (self.p.R / self.p.S) * w1_v[0]
            Gw2_n      = G2_n - (self.p.R / self.p.S) * w2_v[0]
            Gw3_n      = G3_n - (self.p.R / self.p.S) * w3_v[0] - (1 - self.p.R) * self.p.g

            Gw1_nm1    = G1_nm1 - (self.p.R / self.p.S) * w1_v[1]
            Gw2_nm1    = G2_nm1 - (self.p.R / self.p.S) * w2_v[1]
            Gw3_nm1    = G3_nm1 - (self.p.R / self.p.S) * w3_v[1] - (1 - self.p.R) * self.p.g

            w1_np1     = ( w1_v[0] + (h/2.0) * (3.0 * Gw1_n - Gw1_nm1) - xi * Sum_w1 ) / \
                            (1.0 + xi * self.beta_jn(0, nn+1))
            w2_np1     = ( w2_v[0] + (h/2.0) * (3.0 * Gw2_n - Gw2_nm1) - xi * Sum_w2 ) / \
                            (1.0 + xi * self.beta_jn(0, nn+1))
            w3_np1     = ( w3_v[0] + (h/2.0) * (3.0 * Gw3_n - Gw3_nm1) - xi * Sum_w3 ) / \
                            (1.0 + xi * self.beta_jn(0, nn+1))



            x_v        = np.append(x_v, x_np1)
            y_v        = np.append(y_v, y_np1)
            z_v        = np.append(z_v, z_np1)
            w1_v       = np.append(w1_np1, w1_v)
            w2_v       = np.append(w2_np1, w2_v)
            w3_v       = np.append(w3_np1, w3_v)

        pos_vec_x      = x_v
        pos_vec_y      = y_v
        pos_vec_z      = z_v
        w_vec          = np.transpose(np.array([np.flip(w1_v), np.flip(w2_v), np.flip(w3_v)]))

        if flag == True:
            self.pos_vec_x = np.copy(pos_vec_x)
            self.pos_vec_y = np.copy(pos_vec_y)
            self.pos_vec_z = np.copy(pos_vec_z)
            self.w_vec     = np.copy(w_vec)

        return pos_vec_x, pos_vec_y, pos_vec_z, w_vec

    def gamma_jn(self, j, n):

        if n == 0:
          gamma = 0.0

        elif n == 1:
            gamma = self.alpha_jn(j, n)

        elif n == 2:
            gamma = self.beta_jn(j, n)

        elif n == 3:
            if j == 0:
                gamma = (68.0/105.0) * np.sqrt(3.0)
            elif j == 1:
                gamma = (6.0/7.0) * np.sqrt(3.0)
            elif j == 2:
                gamma = (12.0/35.0) * np.sqrt(3.0)
            elif j == 3:
                gamma = (16.0/105.0) * np.sqrt(3.0)

        elif n == 4:
            if j == 0:
                gamma = (244.0/315.0) * np.sqrt(2.0)
            elif j == 1:
                gamma = (1888.0 - 976.0 * np.sqrt(2.0)) / 315.0
            elif j == 2:
                gamma = (488.0 * np.sqrt(2.0) - 656.0 ) / 105.0
            elif j == 3:
                gamma = (544.0/105.0) - (976.0/315.0) * np.sqrt(2.0)
            elif j == 4:
                gamma = (244.0 * np.sqrt(2.0) - 292.0 ) / 315.0

        elif n == 5:
            if j == 0:
                gamma = (244.0/315.0) * np.sqrt(2.0)
            elif j == 1:
                gamma = (362.0/105.0) * np.sqrt(3.0) - (976.0/315.0) * np.sqrt(2.0)
            elif j == 2:
                gamma = (500.0/63.0) * np.sqrt(5.0) - \
                            (1448.0/105.0) * np.sqrt(3.0) + \
                                (488.0/105.0) * np.sqrt(2.0)
            elif j == 3:
                gamma = (-290.0/21.0) * np.sqrt(5.0) + \
                            (724.0/35.0) * np.sqrt(3.0) - \
                                (976.0/315.0) * np.sqrt(2.0)
            elif j == 4:
                gamma = (220.0/21.0) * np.sqrt(5.0) - \
                            (1448.0/105.0) * np.sqrt(3.0) + \
                                (244.0/315.0) * np.sqrt(2.0)
            elif j == 5:
                gamma = (362.0/105.0) * np.sqrt(3.0) - \
                            (164.0/63.0) * np.sqrt(5.0)

        elif n == 6:
            if j == 0:
                gamma = (244.0/315.0) * np.sqrt(2.0)
            elif j == 1:
                gamma = (362.0/105.0) * np.sqrt(3.0) - \
                            (976.0/315.0) * np.sqrt(2.0)
            elif j == 2:
                gamma = (5584.0/315.0) - \
                            (1448.0/105.0) * np.sqrt(3.0) + \
                                (488.0/105.0) * np.sqrt(2.0)
            elif j == 3:
                gamma = (344.0/21.0) * np.sqrt(6.0) - \
                            (22336.0/315.0) + (724.0/35.0) * np.sqrt(3.0) - \
                                (976.0/315.0) * np.sqrt(2.0)
            elif j == 4:
                gamma = (-1188.0/35.0) * np.sqrt(6.0) + \
                            (11168.0/105.0) - (1448.0/105.0) * np.sqrt(3.0) + \
                                (244.0/315.0) * np.sqrt(2.0)
            elif j == 5:
                gamma = (936.0/35.0) * np.sqrt(6.0) - \
                            (22336.0/315.0) + (362.0/105.0) * np.sqrt(3.0)
            elif j == 6:
                gamma = (5584.0/315.0) - (754.0/105.0) * np.sqrt(6.0)

        else:
            if j == 0:
                gamma = 244.0 * np.sqrt(2.0) / 315.0

            elif j == 1:
                gamma = (362.0/105.0) * np.sqrt(3.0) - \
                            (976.0/315.0) * np.sqrt(2.0)

            elif j == 2:
                gamma = (5584.0/315.0) - (1448.0/105.0) * np.sqrt(3.0) + \
                            (488.0/105.0) * np.sqrt(2.0)

            elif j == 3:
                gamma = (1130.0/63.0) * np.sqrt(5.0) - \
                            (22336.0/315.0) + (724.0/35.0) * np.sqrt(3.0) - \
                                (976.0/315.0) * np.sqrt(2.0)

            elif j == n-3:
                gamma = (16.0/105.0) * (n**(7.0/2.0) - \
                            4.0 * (n - 2.0)**(7.0/2.0) + \
                            6.0 * (n - 3.0)**(7.0/2.0) - \
                            4.0 * (n - 4.0)**(7.0/2.0) + \
                                    (n - 5.0)**(7.0/2.0)) - \
                                (8.0/15.0) * n**(5.0/2.0) + \
                                (4.0/9.0) * n**(3.0/2.0) + \
                        (8.0/9.0) * (n - 2.0)**(3.0/2.0) - \
                        (4.0/3.0) * (n - 3.0)**(3.0/2.0) + \
                        (8.0/9.0) * (n - 4.0)**(3.0/2.0) - \
                        (2.0/9.0) * (n - 5.0)**(3.0/2.0)
            elif j == n-2:
                gamma = (16.0/105.0) * ((n - 4.0)**(7.0/2.0) - \
                                    4.0 * (n - 3.0)**(7.0/2.0) + \
                                    6.0 * (n - 2.0)**(7.0/2.0) - \
                                        3.0 * n ** (7.0/2.0)) + \
                                    (32.0/15.0) * n**(5.0/2.0) - \
                                            2.0 * n**(3.0/2.0) - \
                            (4.0/3.0) * (n - 2.0)**(3.0/2.0) + \
                            (8.0/9.0) * (n - 3.0)**(3.0/2.0) - \
                            (2.0/9.0) * (n - 4.0)**(3.0/2.0)
            elif j == n-1:
                gamma = (16.0/105.0) * ( 3.0 * n**(7.0/2.0) - \
                                4.0 * (n - 2.0)**(7.0/2.0) + \
                                        (n - 3.0)**(7.0/2.0)) - \
                                    (8.0/3.0) * n**(5.0/2.0) + \
                                        4.0 * n **(3.0/2.0) + \
                            (8.0/9.0) * (n - 2.0)**(3.0/2.0) - \
                            (2.0/9.0) * (n - 3.0)**(3.0/2.0)
            elif j == n:
                gamma = (16.0/105.0) * ((n - 2.0)**(7.0/2.0) - \
                                                n**(7.0/2.0)) + \
                                    (16.0/15.0) * n**(5.0/2.0) - \
                                    (22.0/9.0) * n**(3.0/2.0) - \
                            (2.0/9.0) * (n - 2.0)**(3.0/2.0) + \
                                    2.0 * np.sqrt(n)
            else:
                gamma = (16.0/105.0) * ((j + 2.0)**(7.0/2.0) + \
                                        (j - 2.0)**(7.0/2.0) - \
                                    4.0 * (j + 1.0)**(7.0/2.0) - \
                                    4.0 * (j - 1.0)**(7.0/2.0) + \
                                            6.0 * j**(7.0/2.0)) + \
                    (2.0/9.0) * (4.0 * (j + 1.0)**(3.0/2.0) + \
                                4.0 * (j - 1.0)**(3.0/2.0) - \
                                        (j + 2.0)**(3.0/2.0) - \
                                        (j - 2.0)**(3.0/2.0) - \
                                        6.0 * j**(3.0/2.0))
        return gamma



    def gamma_v(self, N):

        GammaSubst_v = np.array([])
        for nn in range(0, N+1):
            GammaSubst_v = np.append(GammaSubst_v, self.gamma_jn(nn+1, N+1) -\
                                                    self.gamma_jn(nn, N))

        return GammaSubst_v



    def calc_gamma_mat(self, N):
        for nn in range(0, N-1):
            if nn == 0:
                gamma_mat    = sp.csr_matrix(self.gamma_v(nn), shape=(1, N))
            else:
                gamma_v      = sp.csr_matrix(self.gamma_v(nn), shape=(1, N))
                gamma_mat    = sp.vstack([gamma_mat, gamma_v])
        self.gamma_mat = gamma_mat



    def AdamBashf3(self, t_v, flag=False):
        assert len(t_v) >= 3, "Time grid cannot be used for time stepping."

        h               = t_v[1] - t_v[0]

        # Obtaining first and second time steps with a 1st order method
        assert self.euler_nodes % 2 == 1, "Please provide an odd number of nodes for the Euler computation of the first steps."
        t_np1           = np.linspace(t_v[0], t_v[2], self.euler_nodes)
        pos_vec_x, pos_vec_y, pos_vec_z, w_vec = self.AdamBashf2(t_np1, flag=False)

        # Calculating the rest of the solution with a 3rd order method
        x_v      = np.array([self.x[0], pos_vec_x[int((self.euler_nodes-1)/2)], pos_vec_x[-1]])
        y_v      = np.array([self.x[1], pos_vec_y[int((self.euler_nodes-1)/2)], pos_vec_y[-1]])
        z_v      = np.array([self.x[2], pos_vec_z[int((self.euler_nodes-1)/2)], pos_vec_z[-1]])


        u10, u20, u30  = self.vel.get_velocity(self.x[0], self.x[1], self.x[2], t_v[0])

        w1_v = np.array([w_vec[-1,0], w_vec[int((self.euler_nodes-1)/2),0], self.w[0]])
        w2_v = np.array([w_vec[-1,1], w_vec[int((self.euler_nodes-1)/2),1], self.w[1]])
        w3_v = np.array([w_vec[-1,2], w_vec[int((self.euler_nodes-1)/2),2], self.w[2]])

        xi    = (self.p.R) * np.sqrt(3 / self.p.S) * np.sqrt(h / np.pi)

        for nn in range(2, len(t_v)-1):
            Sum_w1        = np.dot(self.gamma_mat.toarray()[nn, :nn+1], w1_v)
            Sum_w2        = np.dot(self.gamma_mat.toarray()[nn, :nn+1], w2_v)
            Sum_w3        = np.dot(self.gamma_mat.toarray()[nn, :nn+1], w3_v)


            u1_n, u2_n, u3_n = self.vel.get_velocity(x_v[nn], y_v[nn], z_v[nn], t_v[nn])
            u1_nm1, u2_nm1, u3_nm1 = self.vel.get_velocity(x_v[nn-1], y_v[nn-1], z_v[nn-1], t_v[nn-1])
            u1_nm2, u2_nm2, u3_nm2 = self.vel.get_velocity(x_v[nn-2], y_v[nn-2], z_v[nn-2], t_v[nn-2])

            x_np1        = x_v[nn] + (h/12.0) * ( 23.0 * (w1_v[0] + u1_n) -\
                                                    16.0 * (w1_v[1] + u1_nm1) +\
                                                    5.0 * (w1_v[2] + u1_nm2))
            y_np1        = y_v[nn] + (h/12.0) * ( 23.0 * (w2_v[0] + u2_n) -\
                                                    16.0 * (w2_v[1] + u2_nm1) +\
                                                    5.0 * (w2_v[2] + u2_nm2))
            z_np1        = z_v[nn] + (h/12.0) * ( 23.0 * (w3_v[0] + u3_n) -\
                                                    16.0 * (w3_v[1] + u3_nm1) +\
                                                    5.0 * (w3_v[2] + u3_nm2))


            G1_n, G2_n, G3_n     = self.calculate_G(w1_v[0],  w2_v[0], w3_v[0],
                                            x_v[nn], y_v[nn], z_v[nn],
                                            t_v[nn])
            G1_nm1, G2_nm1, G3_nm1 = self.calculate_G(w1_v[1],  w2_v[1], w3_v[1],
                                            x_v[nn-1], y_v[nn-1], z_v[nn-1],
                                            t_v[nn-1])
            G1_nm2, G2_nm2, G3_nm2 = self.calculate_G(w1_v[2],  w2_v[2], w3_v[2],
                                            x_v[nn-2], y_v[nn-2], z_v[nn-2],
                                            t_v[nn-2])


            Gw1_n      = G1_n   - (self.p.R / self.p.S) * w1_v[0]
            Gw2_n      = G2_n   - (self.p.R / self.p.S) * w2_v[0]
            Gw3_n      = G3_n   - (self.p.R / self.p.S) * w3_v[0] - (1 - self.p.R) * self.p.g

            Gw1_nm1    = G1_nm1 - (self.p.R / self.p.S) * w1_v[1]
            Gw2_nm1    = G2_nm1 - (self.p.R / self.p.S) * w2_v[1]
            Gw3_nm1    = G3_nm1 - (self.p.R / self.p.S) * w3_v[1] - (1 - self.p.R) * self.p.g

            Gw1_nm2    = G1_nm2 - (self.p.R / self.p.S) * w1_v[2]
            Gw2_nm2    = G2_nm2 - (self.p.R / self.p.S) * w2_v[2]
            Gw3_nm2    = G3_nm2 - (self.p.R / self.p.S) * w3_v[2] - (1 - self.p.R) * self.p.g



            w1_np1        = ( w1_v[0] + (h/12.0) * (23.0 * Gw1_n - \
                                16.0 * Gw1_nm1 + 5.0 * Gw1_nm2) - \
                                    xi * Sum_w1 ) / \
                            (1.0 + xi * self.gamma_jn(0, nn+1))
            w2_np1        = ( w2_v[0] + (h/12.0) * (23.0 * Gw2_n - \
                                16.0 * Gw2_nm1 + 5.0 * Gw2_nm2) - \
                                    xi * Sum_w2 ) / \
                            (1.0 + xi * self.gamma_jn(0, nn+1))
            w3_np1        = ( w3_v[0] + (h/12.0) * (23.0 * Gw3_n - \
                                16.0 * Gw3_nm1 + 5.0 * Gw3_nm2) - \
                                    xi * Sum_w3 ) / \
                            (1.0 + xi * self.gamma_jn(0, nn+1))


            x_v          = np.append(x_v, x_np1)
            y_v          = np.append(y_v, y_np1)
            z_v          = np.append(z_v, z_np1)
            w1_v         = np.append(w1_np1, w1_v)
            w2_v         = np.append(w2_np1, w2_v)
            w3_v         = np.append(w3_np1, w3_v)

        pos_vec_x      = x_v
        pos_vec_y      = y_v
        pos_vec_z      = z_v
        w_vec          = np.transpose(np.array([np.flip(w1_v), np.flip(w2_v), np.flip(w3_v)]))

        if flag == True:
            self.pos_vec_x = np.copy(pos_vec_x)
            self.pos_vec_y = np.copy(pos_vec_y)
            self.pos_vec_z = np.copy(pos_vec_z)
            self.w_vec     = np.copy(w_vec)

        return pos_vec_x, pos_vec_y, pos_vec_z, w_vec



"""
The 3D velocity field here is (-y, x, 0). This is just the 2D vortex taken along z direction.
"""
import math

class VelocityField3D:

  def __init__(self, omega):
    self.omega    = omega
    self.limits   = False
    self.periodic = False

  def get_velocity(self, x, y, z, t):
    u1 =  -y * self.omega
    u2 =  x * self.omega
    u3 =  0.0 * self.omega
    return u1, u2, u3

  def get_gradient(self, x, y, z, t):
    # TODO : why is there no dependency on x, y, z and t ?
    u1x =  0.0
    u1y =  -1.0 * self.omega
    u1z =  0.0
    u2x =  1.0 * self.omega
    u2y =  0.0
    u2z =  0.0
    u3x =  0.0
    u3y =  0.0
    u3z =  0.0 * self.omega
    return u1x, u1y, u1z, u2x, u2y, u2z, u3x, u3y, u3z

  def get_dudt(self, x, y, z, t):
    # TODO : why is there no dependency on x, y, z and t ?
    u1t = 0.0
    u2t = 0.0
    u3t = 0.0
    return u1t, u2t, u3t


class velocity_field_3d_nondim():

  def __init__(self, omega, char_vel, char_time):
    self.char_vel  = char_vel
    self.char_time = char_time
    self.char_len = char_vel
    self.omega    = omega*char_time
    self.limits   = False
    self.periodic = False

  def get_velocity(self, x, y, z, t):
    u1 =  -y * self.omega*self.char_len
    u2 =  x * self.omega*self.char_len
    u3 =  0.0 * self.omega*self.char_len
    return u1/self.char_vel, u2/self.char_vel, u3/self.char_vel

  def get_gradient(self, x, y, z, t):
    # TODO : why is there no dependency on x, y, z and t ?
    u1x =  0.0
    u1y =  -1.0 * self.omega
    u1z =  0.0
    u2x =  1.0 * self.omega
    u2y =  0.0
    u2z =  0.0
    u3x =  0.0
    u3y =  0.0
    u3z =  0.0 * self.omega
    return u1x*self.char_time, u1y*self.char_time, u1z*self.char_time, u2x*self.char_time, u2y*self.char_time, u2z*self.char_time, u3x*self.char_time, u3y*self.char_time, u3z*self.char_time

  def get_dudt(self, x, y, z, t):
    # TODO : why is there no dependency on x, y, z and t ?
    u1t = 0.0
    u2t = 0.0
    u3t = 0.0
    return u1t, u2t, u3t

class velocity_field_3d_oscillatory():

  def __init__(self):

    self.limits   = False
    self.periodic = False

  def get_velocity(self, x, y, z, t):
    u1 = 0.05
    u2 = math.sin(6*t)
    u3 = 0.0
    return u1, u2, u3

  def get_gradient(self, x, y, z, t):
    u1x =  0.0
    u1y =  0.0
    u1z =  0.0
    u2x =  0.0
    u2y =  0.0
    u2z =  0.0
    u3x =  0.0
    u3y =  0.0
    u3z =  0.0
    return u1x, u1y, u1z, u2x, u2y, u2z, u3x, u3y, u3z

  def get_dudt(self, x, y, z, t):
    u1t = 0.0
    u2t = 6*math.cos(6*t)
    u3t = 0.0
    return u1t, u2t, u3t



"""
TODO : documentation ...
"""
import numpy as np
from scipy.special import erfc, erfcx

from marge3d.params import DaitcheParameters

class AnalyticalSolver:

  def __init__(self, x, v, particle_density, fluid_density, particle_radius,
               kinematic_viscosity, time_scale, char_vel):
      #self.tag       = tag               # particle name/tag
      self.x         = np.copy(x)        # particle position
      self.v         = np.copy(v)        # particle velocity
      #self.vel       = velocity_field
      self.p         = DaitcheParameters(particle_density, fluid_density,
                                    particle_radius, kinematic_viscosity,
                                    time_scale, char_vel)
      #print(time_scale)
      omega          = 10
      Fr             = (0.04*omega**2)/9.8
      self.pseud_St  = (particle_radius**2.0) * omega / (9.0 * kinematic_viscosity )
      #self.time      = tini
      self.pos_vec   = np.copy(x)
      self.vel_vec   = np.copy(v)

      A_cnst         = 1.0 / ((2.0 * self.p.beta + 1.0) * self.pseud_St)
      B_cnst         = (3.0 - (1j / self.pseud_St)) / (2.0 * self.p.beta + 1.0)
      C_cnst         = -3.0 / ((2.0 * self.p.beta + 1.0) * \
                                np.sqrt(np.pi * self.pseud_St))
      #D_cnst         = 2.0 * (1 - self.p.beta) / ((2 * self.p.beta + 1) * self.p.g)
      D_cnst         = 2.0 * (1 - self.p.beta) / ((2 * self.p.beta + 1) * Fr)

      Y1             = self.solve_poly(A_cnst, B_cnst, C_cnst)

      P1             = 0.5 * (-C_cnst * np.sqrt(np.pi) - \
                        np.sqrt(C_cnst**2.0 * np.pi - 4.0 * A_cnst + 4.0 * Y1))
      P2             = 0.5 * (-C_cnst * np.sqrt(np.pi) + \
                        np.sqrt(C_cnst**2.0 * np.pi - 4.0 * A_cnst + 4.0 * Y1))
      Q1             = 0.5 * ( Y1 - np.sqrt( Y1**2.0 - 4.0 * B_cnst ))
      Q2             = 0.5 * ( Y1 + np.sqrt( Y1**2.0 - 4.0 * B_cnst ))

      X1             = 0.5 * (-P1 + np.sqrt(P1**2.0 - 4.0 * Q1))
      X2             = 0.5 * (-P1 - np.sqrt(P1**2.0 - 4.0 * Q1))
      X3             = 0.5 * (-P2 + np.sqrt(P2**2.0 - 4.0 * Q2))
      X4             = 0.5 * (-P2 - np.sqrt(P2**2.0 - 4.0 * Q2))

      self.X         = np.array([X1, X2, X3, X4])


      Z0             = x[0] + x[1] * 1j
      U0             = v[0] + v[1] * 1j

      den            = np.array([])
      for ii in range(0, 4):
          product      = 1.0
          for jj in range(0, 4):
              if ii != jj:
                  product    *= self.X[ii] - self.X[jj]
          assert product != 0.0, "Product of elements in denominator equal to zero."
          den          = np.append(den, product)

      self.A         = np.array([])
      sumAi          = 0.0
      sumAi_Xi       = 0.0
      sumAiXi        = 0.0
      for ii in range(0, 4):
          Ai         = (U0 * (self.X[ii]**2.0 - C_cnst * np.sqrt(np.pi) * self.X[ii]) - \
                        B_cnst * Z0) / den[ii]
          self.A     = np.append(self.A, Ai)
          sumAi     += Ai
          sumAi_Xi  += Ai / self.X[ii]
          sumAiXi   += Ai * self.X[ii]

      if abs(sumAi.real) < 1e-14 and abs(sumAi.imag) < 1e-14:
           sumAi      = 0.0

      assert abs(sumAi.real) < 1e-13 or \
              abs(sumAi.imag) < 1e-13,"Sum of A_i/X_i 's must be equal to initial position"

      assert abs(sumAi_Xi.real - self.x[0]) < 1e-13 or \
              abs(sumAi_Xi.imag - self.x[1]) < 1e-13,"Sum of A_i/X_i 's must be equal to initial position"

      assert abs(sumAiXi.real - self.v[0]) < 1e-13 or \
              abs(sumAiXi.imag - self.v[1]) < 1e-13,"Sum of A_i*X_i 's must be equal to initial velocity"

      #if self.vel.limits == True:
       #   if (self.x[0] > self.vel.x_right or self.x[0] < self.vel.x_left or self.x[1] > self.vel.y_up or self.x[1] < self.vel.y_down):
        #      raise Exception("Particle's initial position is outside the spatial domain")


      V1             = (C_cnst * np.sqrt(np.pi) + np.sqrt(C_cnst**2.0 * np.pi - 4.0 * A_cnst)) / 2.0
      V2             = (C_cnst * np.sqrt(np.pi) - np.sqrt(C_cnst**2.0 * np.pi - 4.0 * A_cnst)) / 2.0

      E1             = D_cnst * ( C_cnst**2 * np.pi - 2*A_cnst - (C_cnst * np.sqrt(np.pi)) * np.sqrt(C_cnst**2.0 * np.pi - 4.0 * A_cnst)) / \
                       (2*A_cnst**2 * np.sqrt(C_cnst**2.0 * np.pi - 4.0 * A_cnst))
      E2             = D_cnst * (-C_cnst**2 * np.pi + 2*A_cnst - (C_cnst * np.sqrt(np.pi)) * np.sqrt(C_cnst**2.0 * np.pi - 4.0 * A_cnst)) / \
                        (2*A_cnst**2 * np.sqrt(C_cnst**2.0 * np.pi - 4.0 * A_cnst))

      Coeff_z1       = 2 * D_cnst * C_cnst / A_cnst**2
      Coeff_z2       = D_cnst/A_cnst

      self.V         = np.array([V1, V2])
      self.E         = np.array([E1, E2])
      self.coeff_z   = np.array([Coeff_z1, Coeff_z2])


  def solve_poly(self, A, B, C):
      p        = np.array([1.0, -A, -(4.0 * B + np.pi * C**2.0 * 1j),
                           4.0 * A * B + C**2.0 * np.pi * (1.0 - B)])

      root_vec = np.roots(p)

      real_max = -np.inf
      for ii in range(0, len(root_vec)):
          if real_max.real < root_vec[ii].real:
              real_max = root_vec[ii]

      return real_max

  def exp_x_erfc(self, z):
      np.seterr(all="raise")
      # Avoid Overflow using log
      result  = z**2.0 + np.log(erfc(z))
      result  = np.exp(result)
      return result

  def solve(self, t):

      pos_result = 0.0
      vel_result = 0.0
      for ii in range(4):
          coeff   = self.A[ii] / self.X[ii]
          #print("X_i: " + str(self.X[ii]))
          pos_result += coeff * erfcx(-self.X[ii] * np.sqrt(t))
          #pos_result += coeff * self.exp_x_erfc(-self.X[ii] * np.sqrt(t))
          #vel_result += self.A[ii] * self.X[ii] * np.exp(t * self.X[ii]**2.0) *\
           #               erfc(-self.X[ii] * np.sqrt(t))

      x            = np.array([pos_result.real, pos_result.imag])
      #v            = np.array([vel_result.real, vel_result.imag])

      #self.pos_vec = np.vstack((self.pos_vec, x))
      #self.vel_vec = np.vstack((self.vel_vec, v))


      return x[0], x[1]



  def solve_z(self, t):

      z1 = (self.coeff_z[0]) * (np.sqrt(t)) + (self.coeff_z[1]) * t + \
           (self.E[0]/self.V[0])*(erfcx(-self.V[0] * np.sqrt(t))-1) + \
           (self.E[1]/self.V[1])*(erfcx(-self.V[1] * np.sqrt(t))-1)
           #(self.E[0]/self.V[0])*(np.exp(self.V[0]**2 * t)* erfc(-self.V[0] * np.sqrt(t))-1) + \
           #(self.E[1]/self.V[1])*(np.exp(self.V[1]**2 * t)* erfc(-self.V[1] * np.sqrt(t))-1)

      #z1 = (self.coeff_z[0]) * (np.sqrt(t)) + (self.coeff_z[1]) * t + \
       #    (self.E[0]/self.V[0])*(self.exp_x_erfc(-self.V[0] * np.sqrt(t))-1) + \
        #   (self.E[1]/self.V[1])*(self.exp_x_erfc(-self.V[1] * np.sqrt(t))-1)

      return z1



"""
List of parameters required for 3D MRE as solved according to Daitche's paper
"""


TEST_PARAMETERS = dict(
    particle_density    = 500,
    fluid_density       = 972,
    particle_radius     = 0.00018**0.5,
    kinematic_viscosity = 2e-4,
    time_scale          = 0.1,
    char_vel            = 0.4,
)


class DaitcheParameters:

  def set_beta(self):
    self.beta = self.rho_p /self.rho_f

  def set_S(self):
    self.S = (1.0/3.0)*self.a**2/(self.nu * self.T)

  def set_R(self):
    self.R = 3.0/(1.0 + 2.0*self.beta)

  def set_gravity(self):
    self.g = (self.T/self.U)*9.8

  def __init__(self, particle_density, fluid_density, particle_radius,
                     kinematic_viscosity, time_scale, char_vel):
    self.rho_p = particle_density
    self.rho_f = fluid_density
    self.a     = particle_radius
    self.nu    = kinematic_viscosity
    self.T     = time_scale
    self.U     = char_vel

    self.set_beta()
    self.set_S()
    self.set_R()
    self.set_gravity()


import numpy as np
import matplotlib.pyplot as plt

from marge3d.numeric import NumericalSolver
from marge3d.fields import VelocityField3D

from marge3d.params import DaitcheParameters

R0     = np.array([1, 0, 0])
W0     = np.array([0, 0, 0])

particle_density    = 1410
fluid_density       = 972
particle_radius     = 0.0015
kinematic_viscosity = 2 * 1e-4
time_scale          = 0.0125
char_vel            = 0.4

par = DaitcheParameters(particle_density, fluid_density, particle_radius,
                    kinematic_viscosity, time_scale, char_vel)

print("G = ", par.g)
print("R = ", par.R)
print("S = ", par.S)

order  = 2
T_ini  = 0
T_fin  = 10
T      = T_fin - T_ini
Vortex = VelocityField3D(1)

V0    = Vortex.get_velocity(R0[0], R0[1], R0[2], T_ini)
N     = 100

Order_n = NumericalSolver(R0, W0, Vortex, N, order, particle_density, fluid_density, particle_radius,
                   kinematic_viscosity, time_scale, char_vel)

t_v = np.linspace(T_ini, T_fin, N)
if order == 1:
    R_x, R_y, R_z, W = Order_n.Euler(t_v, flag=True)
elif order == 2:
    R_x, R_y, R_z, W = Order_n.AdamBashf2(t_v, flag=True)
elif order == 3:
    R_x, R_y, R_z, W = Order_n.AdamBashf3(t_v, flag=True)
else:
    raise ValueError("Order must be 1, 2, or 3.")

#"""
# For quiver plot of the fluid field
x, y, z = np.meshgrid(
    np.linspace(-1.5, 1.5, 5),
    np.linspace(-1.5, 1.5, 5),
    np.linspace(-0.21, 0, 5)
)

u, v, w = Vortex.get_velocity(x, y, z, T_fin)

fig = plt.figure()
# size of the figure
fig.set_size_inches(5, 5)
ax = fig.add_subplot(111, projection='3d')

# Plot the 3D line
plt.plot(R_x, R_y, R_z, color="blue", label='Numerical solution')
ax.scatter(1, 0, 0, color='green', label='Initial position')
ax.quiver(x, y, z, u, v, w, length=0.25, normalize=True, arrow_length_ratio=0.1, alpha=0.5)

# Add labels
ax.set_xlabel('x')
ax.set_ylabel('y')
ax.set_zlabel('z')

# Add a legend
ax.legend(loc='upper right', bbox_to_anchor=(1, 0.9))

# Show the plot
ax.set_box_aspect(None, zoom=0.8)

plt.show()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Analytic solution and numerical solution will agree only when
time_scale is 0.1 and char_vel is 0.4.
This is because the analytical solution is derived under these conditions.
Other parameters can be changed.
"""
import numpy as np
import matplotlib.pyplot as plt

from marge3d.fields import VelocityField3D
from marge3d.numeric import NumericalSolver
from marge3d.analytic import AnalyticalSolver


particle_density    = 500
fluid_density       = 972
particle_radius     = 0.0015 # S=0.3 for this and the 3rd order method does not perform well. So the solution trajectories are obtained for order 2 method.
kinematic_viscosity = 2 * 1e-4
time_scale          = 0.1
char_vel            = 0.4

order  = 2
T      = 10
h      = 0.01 #Somehow if I take h<0.01, the numerical solution does not converge, it becomes unstable.
#This happens only when we use the characteristic velocity=0.4 and time scale=0.1.
N      = 1000

R0     = np.array([1, 0, 0])
W0     = np.array([0, 0, 0])

Vortex = VelocityField3D(10*time_scale)
t_v   = np.linspace(0, T, N)
#t_v    = np.arange(0, T, h)

U0    = Vortex.get_velocity(R0[0], R0[1], R0[2], 0) # Initial fluid velocity

MRE_analytic = AnalyticalSolver(R0, U0, particle_density, fluid_density, particle_radius, kinematic_viscosity, time_scale, char_vel)

Order_n = NumericalSolver(R0, W0, Vortex, len(t_v), order, particle_density, fluid_density, particle_radius,
                   kinematic_viscosity, time_scale, char_vel)

if order == 1:
    R_x, R_y, R_z, W = Order_n.Euler(t_v, flag=True)
elif order == 2:
    R_x, R_y, R_z, W = Order_n.AdamBashf2(t_v, flag=True)
elif order == 3:
    R_x, R_y, R_z, W = Order_n.AdamBashf3(t_v, flag=True)
else:
    raise ValueError("Order must be 1, 2, or 3.")

X      = np.zeros(len(t_v))
Y      = np.zeros(len(t_v))
Z      = np.zeros(len(t_v))

for i in range(len(t_v)):
    X[i], Y[i] = MRE_analytic.solve(t_v[i])
    Z[i]       = MRE_analytic.solve_z(t_v[i])

#"""
# For quiver plot of the fluid field
x, y, z = np.meshgrid(
    np.linspace(-1.5, 1.5, 5),
    np.linspace(-1.5, 1.5, 5),
    np.linspace(0, 0.27, 5)
)
#"""

u, v, w = Vortex.get_velocity(x, y, z, T)

#'''
fig = plt.figure()
fig.set_size_inches(5, 5)
ax = fig.add_subplot(111, projection='3d')

# Plot the 3D line
plt.plot(X, Y, Z, label="Analytical solution", color="red")
#plt.plot(R_x, R_y, R_z, label="Numerical solution", color="blue")
plt.plot(R_x[0::50], R_y[0::50], R_z[0::50], 'x', label = "Numerical solution", color="blue")
ax.scatter(1, 0, 0, color='green', label='Initial position')
ax.quiver(x, y, z, u, v, w, length=0.25, normalize=True, arrow_length_ratio=0.1, alpha=0.5)

# Add labels
ax.set_xlabel('x')
ax.set_ylabel('y')
ax.set_zlabel('z')

# Add a legend
ax.legend(loc='upper right', bbox_to_anchor=(1, 0.9))

ax.set_box_aspect(None, zoom=0.8)

# Show the plot
plt.show()


import numpy as np

def numerical_order(n_steps, err):
    """
    Compute numerical order from two vectors containing the error and the number of time-steps.

    Parameters
    ----------
    n_steps : np.1darray or list
        Different number of steps to compute the error.
    err : np.1darray
        Different error values associated to the number of steps.

    Returns
    -------
    beta : float
        Order coefficient computed through linear regression.
    rmse : float
        The root mean square error of the linear regression.
    """
    n_steps = np.asarray(n_steps)
    x, y = np.log10(1/n_steps), np.log10(err)

    # Compute regression coefficients and rmse
    xMean = x.mean()
    yMean = y.mean()
    sX = ((x-xMean)**2).sum()
    sXY = ((x-xMean)*(y-yMean)).sum()

    beta = sXY/sX
    alpha = yMean - beta*xMean

    yHat = alpha + beta*x
    rmse = ((y-yHat)**2).sum()**0.5
    rmse /= x.size**0.5

    return beta, rmse


"""
###############################################################################
################################Convergence test###############################
###############################################################################
Particle initial velocity is set to be equal to the initial fluid velocity.
This is according to the paper by Candelier et al.
"""
import numpy as np
import matplotlib.pyplot as plt
from concurrent.futures import ProcessPoolExecutor

from marge3d.fields import VelocityField3D
from marge3d.numeric import NumericalSolver
from marge3d.analytic import AnalyticalSolver

# -----------------------------------------------------------------------------
# Parameters
# -----------------------------------------------------------------------------
particle_density    = 500
fluid_density       = 972
particle_radius     = 0.00018**0.5  # 0.0135
kinematic_viscosity = 2 * 1e-4
time_scale          = 0.1
char_vel            = 0.4

St = (particle_radius**2.0) * 1 / (3 * kinematic_viscosity*time_scale)
print("Stokes number (S) = ", St)

# -----------------------------------------------------------------------------
# Initial conditions
# -----------------------------------------------------------------------------
R0     = np.array([1, 0, 0])
Vortex = VelocityField3D(1)
U0     = Vortex.get_velocity(R0[0], R0[1], R0[2], 0) # Initial fluid velocity
V0     = np.array(Vortex.get_velocity(R0[0], R0[1], R0[2], 0)) # Same as initial fluid velocity
W0     = V0 - U0 # Relative velocity

MRE_analytic = AnalyticalSolver(R0, U0, particle_density, fluid_density,
                                       particle_radius, kinematic_viscosity, time_scale, char_vel)

# -----------------------------------------------------------------------------
# Helper functions
# -----------------------------------------------------------------------------
def analytic_traj(t):
    """Analytical position at time t."""
    sol = MRE_analytic.solve(t)
    return np.array([sol[0], sol[1], MRE_analytic.solve_z(t)])

def Err_n(t, N, order):
    """Compute L_inf error for given N and order."""
    t_v = np.linspace(0, t, N)
    mre = NumericalSolver(R0, W0, Vortex, N+1, order, particle_density, fluid_density, particle_radius,
                       kinematic_viscosity, time_scale, char_vel)
    if order == 1:
        R_x, R_y, R_z, W = mre.Euler(t_v, flag=True)
    elif order == 2:
        R_x, R_y, R_z, W = mre.AdamBashf2(t_v, flag=True)
    elif order == 3:
        R_x, R_y, R_z, W = mre.AdamBashf3(t_v, flag=True)
    else:
        raise ValueError("Order must be 1, 2, or 3")

    R_num = np.column_stack((R_x, R_y, R_z))
    R_ana = np.array([analytic_traj(ti) for ti in t_v])
    return np.max(np.linalg.norm(R_ana - R_num, axis=1))


# -----------------------------------------------------------------------------
# Compute errors
# -----------------------------------------------------------------------------
def _err_worker(args):
    """Worker function for parallel error computation."""
    return Err_n(*args)

def compute_all_errors(N_v, t_end=1.0):
    with ProcessPoolExecutor() as ex:
        results = list(ex.map(
            _err_worker,
            [(t_end, N, order) for N in N_v for order in [1, 2, 3]]
        ))
    results = np.array(results).reshape(len(N_v), 3)
    return (*results.T,)


# -----------------------------------------------------------------------------
# Main script
# -----------------------------------------------------------------------------
if __name__ == "__main__":
    N_v = np.logspace(np.log10(10), np.log10(1000), num=20).astype(int)
    Err_v1, Err_v2, Err_v3 = compute_all_errors(N_v)

    plt.figure(figsize=(5.5, 4))
    lw, ms = 2.0, 6
    colors = ["blue", "orange", "green"]

    for Err, order, color in zip([Err_v1, Err_v2, Err_v3],
                                [1, 2, 3], colors):
        plt.plot(N_v, Err, 'o-', color=color, lw=lw, ms=ms,
                label=f"{order}st order" if order==1 else f"{order}nd order" if order==2 else "3rd order")

    # ---------------------------------------------------------------------
    # reference slopes
    # ---------------------------------------------------------------------
    ref_x = np.array([50, 130])  # short horizontal range
    ref_y_levels = [1e-1, 1e-3, 1e-5]

    for order, y0 in zip([1, 2, 3], ref_y_levels):
        ref_y = y0 * (ref_x / ref_x[0]) ** (-order)
        plt.plot(ref_x, ref_y, '--', color='gray', lw=1.5, alpha=0.7)
        plt.text(ref_x[0] * 1.6, ref_y[0], f"$N^{{-{order}}}$",
                fontsize=11, color='gray', rotation=0)

    # ---------------------------------------------------------------------
    plt.xscale("log"); plt.yscale("log")
    plt.xlabel(r"Nodes $N$", fontsize=13)
    plt.ylabel(r"$L_{\infty}$ error", fontsize=13)
    plt.xticks(fontsize=11); plt.yticks(fontsize=11)
    plt.legend(fontsize=11, frameon=True, loc="lower left", framealpha=0.9)
    plt.grid()
    plt.tight_layout()
    plt.show()



import pytest

import numpy as np

from marge3d.fields import VelocityField3D
from marge3d.analytic import AnalyticalSolver
from marge3d.numeric import NumericalSolver
from marge3d.params import TEST_PARAMETERS
from marge3d.utils import numerical_order

N_STEPS_CONV = [100, 200, 500]

@pytest.mark.parametrize("order", [1, 2, 3])
@pytest.mark.parametrize("t_end", [1.0, 4.0])
def test_order(t_end, order):

    R0     = np.array([1, 0, 0])
    vortex = VelocityField3D(1)
    U0     = vortex.get_velocity(R0[0], R0[1], R0[2], 0) # Initial fluid velocity
    V0     = np.array(vortex.get_velocity(R0[0], R0[1], R0[2], 0)) # Same as initial fluid velocity
    W0     = V0 - U0 # Relative velocity
    analytic = AnalyticalSolver(R0, U0, **TEST_PARAMETERS)

    errors = []
    for n in N_STEPS_CONV:
        times = np.linspace(0, t_end, n)
        solver = NumericalSolver(R0, W0, vortex, n+1, order, **TEST_PARAMETERS)

        stepper = {
            1: solver.Euler,
            2: solver.AdamBashf2,
            3: solver.AdamBashf3,
        }[order]

        R_x, R_y, R_z, _ = stepper(times, flag=True)
        R_num = np.column_stack((R_x, R_y, R_z))
        R_ana = np.array([[*analytic.solve(t), analytic.solve_z(t)] for t in times])

        err = np.max(np.linalg.norm(R_ana - R_num, axis=1))
        errors.append(err)

    num_order, rmse = numerical_order(N_STEPS_CONV, errors)
    expected_order = order
    if order == 3:
        # TODO : this is weird ... implementation should be checked
        expected_order = 2.5

    assert rmse < 0.05, \
        f"rmse to high ({rmse}) for {order=}"
    assert abs(num_order-expected_order) < 0.1, \
        f"expected order {expected_order:.2f}, but got {num_order:.2f} for {order=}"
